package cc.fascinated.backend.model.leaderboard;

import cc.fascinated.backend.common.DateUtils;
import cc.fascinated.backend.model.score.Score;
import cc.fascinated.backend.model.token.ScoreSaberLeaderboardToken;
import lombok.AllArgsConstructor;
import lombok.Getter;
import org.springframework.data.mongodb.core.mapping.Document;

import java.util.Date;

/**
 * A leaderboard for a song.
 */
@Document
@AllArgsConstructor @Getter
public class Leaderboard {
    /**
     * The ID of the leaderboard.
     */
    private final String id;

    /**
     * The hash of the song.
     */
    private final String songHash;

    /**
     * The name of the song.
     */
    private String songName;

    /**
     * The sub name of the song.
     */
    private String songSubName;

    /**
     * The author of the song.
     */
    private String songAuthorName;

    /**
     * The mapper of the song.
     */
    private String levelAuthorName;

    /**
     * The difficulty of the song.
     */
    private Difficulty difficulty;

    /**
     * The maximum score of the song.
     */
    private int maxScore;

    /**
     * The date the leaderboard was created.
     */
    private Date createdDate;

    /**
     * The date the song was ranked.
     */
    private Date rankedDate;

    /**
     * The date the song was qualified.
     */
    private Date qualifiedDate;

    /**
     * The date the song's status was changed to loved.
     */
    private Date lovedDate;

    /**
     * Whether this leaderboard is ranked.
     */
    private boolean ranked;

    /**
     * Whether this leaderboard is qualified to be ranked.
     */
    private boolean qualified;

    /**
     * Whether this leaderboard is in a loved state.
     */
    private boolean loved;

    /**
     * The maximum PP for this leaderboard.
     */
    private int maxPP;

    /**
     * The star rating for this leaderboard.
     */
    private double stars;

    /**
     * The amount of plays for this leaderboard.
     */
    private int plays;

    /**
     * The amount of daily plays for this leaderboard.
     */
    private int dailyPlays;

    /**
     * Whether this leaderboard has positive modifiers.
     */
    private boolean positiveModifiers;

    /**
     * The cover image for this leaderboard.
     */
    private String coverImage;

    /**
     * Gets the leaderboard from a leaderboard token.
     *
     * @param token The token.
     * @return The leaderboard.
     */
    public static Leaderboard fromToken(ScoreSaberLeaderboardToken token) {
        return new Leaderboard(
                token.getId(),
                token.getSongHash(),
                token.getSongName(),
                token.getSongSubName(),
                token.getSongAuthorName(),
                token.getLevelAuthorName(),
                Difficulty.fromToken(token.getDifficulty()),
                token.getMaxScore(),
                token.getCreatedDate() == null ? null : DateUtils.getDateFromString(token.getCreatedDate()),
                token.getRankedDate() == null ? null : DateUtils.getDateFromString(token.getRankedDate()),
                token.getQualifiedDate() == null ? null : DateUtils.getDateFromString(token.getQualifiedDate()),
                token.getLovedDate() == null ? null : DateUtils.getDateFromString(token.getLovedDate()),
                token.isRanked(),
                token.isQualified(),
                token.isLoved(),
                token.getMaxPP(),
                token.getStars(),
                token.getPlays(),
                token.getDailyPlays(),
                token.isPositiveModifiers(),
                token.getCoverImage()
        );
    }

    /**
     * A difficulty for a leaderboard.
     */
    @AllArgsConstructor @Getter
    public static class Difficulty {
        /**
         * The ID of the difficulty.
         */
        private final int id;

        /**
         * The name of the difficulty.
         */
        private final Score.Difficulty difficulty;

        /**
         * The raw name of the difficulty.
         */
        private final String rawDifficulty;

        /**
         * The gamemode of the difficulty.
         */
        private final String gamemode;

        /**
         * Gets the difficulty from a token.
         *
         * @param token The token.
         * @return The difficulty.
         */
        public static Difficulty fromToken(ScoreSaberLeaderboardToken.Difficulty token) {
            return new Difficulty(
                    token.getLeaderboardId(),
                    Score.Difficulty.fromId(token.getDifficulty()),
                    token.getDifficultyRaw(),
                    token.getGameMode()
            );
        }
    }
}
